//  Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.

import 'dart:math';

import 'package:flutter/foundation.dart';
import 'package:flutter/services.dart';

import 'constants/channel_constants.dart';
import 'constants/key_constants.dart';
import 'constants/method_constants.dart';
import 'enums/zone_query_policy.dart';
import 'exceptions/exception.dart';
import 'models/zone_config.dart';
import 'models/zone_snapshot.dart';
import 'query.dart';
import 'transaction.dart';

class AGConnectCloudDBZone {
  final String _id;
  final MethodChannel _methodChannel = const MethodChannel(ChannelConstants.METHOD_CHANNEL);
  final List<String> _activeSnapshotSubscriptionIds = <String>[];

  AGConnectCloudDBZone._internal(this._id);

  factory AGConnectCloudDBZone.fromMap(Map<String, dynamic> map) {
    return AGConnectCloudDBZone._internal(map[KeyConstants.ZONE_ID]);
  }

  Map<String, dynamic> toMap() {
    return <String, dynamic>{
      KeyConstants.ZONE_ID: _id,
    };
  }

  /// You can invoke this method to obtain the current Cloud DB zone configuration information,
  /// including the Cloud DB zone name, synchronization attributes, and access attributes.
  Future<AGConnectCloudDBZoneConfig> getCloudDBZoneConfig() async {
    try {
      final Map<dynamic, dynamic> result = await _methodChannel.invokeMethod<Map<dynamic, dynamic>>(
        MethodConstants.GET_CLOUD_DB_ZONE_CONFIG,
        <String, dynamic>{
          KeyConstants.ZONE_ID: _id,
        },
      );
      return AGConnectCloudDBZoneConfig.fromMap(Map<String, dynamic>.from(result));
    } catch (e) {
      throw AGConnectCloudDBException.from(e);
    }
  }

  /// You can invoke this method to write a group of objects to the current Cloud DB zone in batches.
  /// The write operation is atomic. That is, the input objects are either
  /// all successfully written or all fail to be written.
  /// If an object with the same primary key exists in the Cloud DB zone,
  /// the existing object data will be updated. Otherwise, a new object is written.
  ///
  /// Returns the number of objects that are successfully written.
  ///
  /// **Comply with the following rules when invoking this method:**
  ///
  /// * You can invoke this method to write object data only when Cloud DB zone is opened.
  ///   Otherwise, the write operation will fail.
  /// * The data size of each object in the list must be less than or equal to 2 MB.
  ///   Otherwise, the object data fails to be written.
  /// * The total number of data records in the list must be less than or equal to 1000.
  ///   Otherwise, the write operation fails.
  /// * The data size of all objects in the list must be less than or equal to 2 MB
  ///   only when the data synchronization attribute is cache and
  ///   the data persistency attribute is non-persistent.
  Future<int> executeUpsert({
    @required String objectTypeName,
    @required List<Map<String, dynamic>> entries,
  }) async {
    if (objectTypeName == null || objectTypeName.isEmpty) {
      throw FormatException('Value of objectTypeName cannot be null or an empty string.', objectTypeName);
    }
    if (entries == null || entries.isEmpty) {
      throw FormatException('Value of entries cannot be null or an empty list.', entries);
    }

    try {
      return await _methodChannel.invokeMethod<int>(
        MethodConstants.EXECUTE_UPSERT,
        <String, dynamic>{
          KeyConstants.ZONE_ID: this._id,
          KeyConstants.ZONE_OBJECT_TYPE_NAME: objectTypeName,
          KeyConstants.ZONE_OBJECT_DATA_ENTRIES: entries,
        },
      );
    } catch (e) {
      throw AGConnectCloudDBException.from(e);
    }
  }

  /// You can call this method to delete a group of Cloud DB zone objects
  /// whose primary key values are the same as those of the input object list in batches.
  /// The delete operation focuses only on whether the primary key of the object is consistent,
  /// regardless of whether other attributes of the object match the stored data.
  /// The delete operation is atomic. That is, the input object list is
  /// either all deleted successfully or all fails to be deleted.
  ///
  /// Returns the number of objects that are successfully deleted.
  ///
  /// **Comply with the following rules when invoking this method:**
  ///
  /// * You can call this method to delete objects only when Cloud DB zone is opened.
  ///   Otherwise, the delete operation fails.
  /// * When a group of objects is deleted, the total number of data records in the list
  ///   must be less than or equal to 1000. Otherwise, the delete operation fails.
  /// * The data size of all objects in the list must be less than or equal to 2 MB.
  ///   Otherwise, the delete operation fails.
  /// * The data size of all objects in the list must be less than or equal to 2 MB
  ///   only when the data synchronization attribute is cache and
  ///   the data persistency attribute is non-persistent.
  Future<int> executeDelete({
    @required String objectTypeName,
    @required List<Map<String, dynamic>> entries,
  }) async {
    if (objectTypeName == null || objectTypeName.isEmpty) {
      throw FormatException('Value of objectTypeName cannot be null or an empty string.', objectTypeName);
    }
    if (entries == null || entries.isEmpty) {
      throw FormatException('Value of entries cannot be null or an empty list.', entries);
    }

    try {
      return await _methodChannel.invokeMethod<int>(
        MethodConstants.EXECUTE_DELETE,
        <String, dynamic>{
          KeyConstants.ZONE_ID: this._id,
          KeyConstants.ZONE_OBJECT_TYPE_NAME: objectTypeName,
          KeyConstants.ZONE_OBJECT_DATA_ENTRIES: entries,
        },
      );
    } catch (e) {
      throw AGConnectCloudDBException.from(e);
    }
  }

  /// You can invoke this method to query object data that meets
  /// specific conditions from the Cloud DB zone.
  ///
  /// **Comply with the following rules when invoking this method:**
  ///
  /// * You can invoke this method to query object data only when Cloud DB zone is opened.
  ///   Otherwise, the query will fail.
  /// * When querying data, ensure that the size of each query result is
  ///   less than or equal to 5 MB. Otherwise, the query fails.
  /// * It is recommended that the number of data records in the query result be
  ///   less than or equal to 100. Otherwise, the performance deteriorates.
  ///   The number of data records in the query result cannot exceed 1000. Otherwise, the query fails.
  Future<AGConnectCloudDBZoneSnapshot> executeQuery({
    @required AGConnectCloudDBQuery query,
    @required AGConnectCloudDBZoneQueryPolicy policy,
  }) async {
    if (query == null) {
      throw FormatException('Value of query cannot be null.', query);
    }
    if (policy == null) {
      throw FormatException('Value of policy cannot be null.', policy);
    }

    try {
      final Map<dynamic, dynamic> result = await _methodChannel.invokeMethod<Map<dynamic, dynamic>>(
        MethodConstants.EXECUTE_QUERY,
        <String, dynamic>{
          KeyConstants.ZONE_ID: this._id,
          KeyConstants.QUERY: query.query,
          KeyConstants.POLICY_INDEX: policy.index,
        },
      );
      return AGConnectCloudDBZoneSnapshot.fromMap(Map<String, dynamic>.from(result));
    } catch (e) {
      throw AGConnectCloudDBException.from(e);
    }
  }

  /// You can invoke this method to query Cloud DB zone object data that
  /// meets specified conditions and return the average value of specified fields.
  ///
  /// **Comply with the following rules when invoking this method:**
  ///
  /// * You can invoke this method to query the average value of specified fields
  ///   only when Cloud DB zone is opened. Otherwise, the query will fail.
  /// * Supports the following data types:
  ///   * **int** for Short, Integer, and Long.
  ///   * **double** for Double, and Float.
  Future<dynamic> executeAverageQuery({
    @required String field,
    @required AGConnectCloudDBQuery query,
    @required AGConnectCloudDBZoneQueryPolicy policy,
  }) async {
    if (field == null || field.isEmpty) {
      throw FormatException('Value of field cannot be null or an empty string.', field);
    }
    if (query == null) {
      throw FormatException('Value of query cannot be null.', query);
    }
    if (policy == null) {
      throw FormatException('Value of policy cannot be null.', policy);
    }

    try {
      // Hint for null-safety: Return value can be null, int or double.
      return await _methodChannel.invokeMethod<dynamic>(
        MethodConstants.EXECUTE_AVERAGE_QUERY,
        <String, dynamic>{
          KeyConstants.ZONE_ID: this._id,
          KeyConstants.FIELD: field,
          KeyConstants.QUERY: query.query,
          KeyConstants.POLICY_INDEX: policy.index,
        },
      );
    } catch (e) {
      throw AGConnectCloudDBException.from(e);
    }
  }

  /// You can call this method to search for Cloud DB zone objects that meet
  /// the specified conditions and return the sum of the data record values of the designated fields.
  ///
  /// **Comply with the following rules when invoking this method:**
  ///
  /// * This method is called to query the sum of the specified fields
  ///   only when Cloud DB zone is opened. Otherwise, the query will fail.
  /// * The startAt(), startAfter(), endAt(), and endBefore() methods cannot be used as query conditions.
  /// * Supports the following data types:
  ///   * **int** for Byte, Short, Integer, and Long.
  ///   * **double** for Double, and Float.
  Future<dynamic> executeSumQuery({
    @required String field,
    @required AGConnectCloudDBQuery query,
    @required AGConnectCloudDBZoneQueryPolicy policy,
  }) async {
    if (field == null || field.isEmpty) {
      throw FormatException('Value of field cannot be null or an empty string.', field);
    }
    if (query == null) {
      throw FormatException('Value of query cannot be null.', query);
    }
    if (policy == null) {
      throw FormatException('Value of policy cannot be null.', policy);
    }

    try {
      // Hint for null-safety: Return value can be null, int or double.
      return await _methodChannel.invokeMethod<dynamic>(
        MethodConstants.EXECUTE_SUM_QUERY,
        <String, dynamic>{
          KeyConstants.ZONE_ID: this._id,
          KeyConstants.FIELD: field,
          KeyConstants.QUERY: query.query,
          KeyConstants.POLICY_INDEX: policy.index,
        },
      );
    } catch (e) {
      throw AGConnectCloudDBException.from(e);
    }
  }

  /// You can call this method to search for Cloud DB zone objects that meet
  /// the specified conditions and return the maximum value of the data record of the specified fields.
  ///
  /// **Comply with the following rules when invoking this method:**
  ///
  /// * This method is called to query the maximum value of the specified fields
  ///   only when Cloud DB zone is opened. Otherwise, the query will fail.
  /// * The startAt(), startAfter(), endAt(), and endBefore() methods cannot be used as query conditions.
  /// * Supports the following data types:
  ///   * **int** for Byte, Short, Integer, and Long.
  ///   * **double** for Double, and Float.
  Future<dynamic> executeMaximumQuery({
    @required String field,
    @required AGConnectCloudDBQuery query,
    @required AGConnectCloudDBZoneQueryPolicy policy,
  }) async {
    if (field == null || field.isEmpty) {
      throw FormatException('Value of field cannot be null or an empty string.', field);
    }
    if (query == null) {
      throw FormatException('Value of query cannot be null.', query);
    }
    if (policy == null) {
      throw FormatException('Value of policy cannot be null.', policy);
    }

    try {
      // Hint for null-safety: Return value can be null, int or double.
      return await _methodChannel.invokeMethod<dynamic>(
        MethodConstants.EXECUTE_MAXIMUM_QUERY,
        <String, dynamic>{
          KeyConstants.ZONE_ID: this._id,
          KeyConstants.FIELD: field,
          KeyConstants.QUERY: query.query,
          KeyConstants.POLICY_INDEX: policy.index,
        },
      );
    } catch (e) {
      throw AGConnectCloudDBException.from(e);
    }
  }

  /// You can call this method to search for Cloud DB zone objects that meet
  /// the specified conditions and return the minimum value of the data record of the specified fields.
  ///
  /// **Comply with the following rules when invoking this method:**
  ///
  /// * This method is called to query the minimum value of the specified fields
  ///   only when Cloud DB zone is opened. Otherwise, the query will fail.
  /// * The startAt(), startAfter(), endAt(), and endBefore() methods cannot be used as query conditions.
  /// * Supports the following data types:
  ///   * **int** for Byte, Short, Integer, and Long.
  ///   * **double** for Double, and Float.
  Future<dynamic> executeMinimalQuery({
    @required String field,
    @required AGConnectCloudDBQuery query,
    @required AGConnectCloudDBZoneQueryPolicy policy,
  }) async {
    if (field == null || field.isEmpty) {
      throw FormatException('Value of field cannot be null or an empty string.', field);
    }
    if (query == null) {
      throw FormatException('Value of query cannot be null.', query);
    }
    if (policy == null) {
      throw FormatException('Value of policy cannot be null.', policy);
    }

    try {
      // Hint for null-safety: Return value can be null, int or double.
      return await _methodChannel.invokeMethod<dynamic>(
        MethodConstants.EXECUTE_MINIMAL_QUERY,
        <String, dynamic>{
          KeyConstants.ZONE_ID: this._id,
          KeyConstants.FIELD: field,
          KeyConstants.QUERY: query.query,
          KeyConstants.POLICY_INDEX: policy.index,
        },
      );
    } catch (e) {
      throw AGConnectCloudDBException.from(e);
    }
  }

  /// You can call this method to search for Cloud DB zone objects that meet
  /// specified conditions and return the number of data records in a specified field.
  /// During statistics collection, if the value of this field is not empty, count 1. Otherwise, count 0.
  ///
  /// **Comply with the following rules when invoking this method:**
  ///
  /// * This method is called to query the number of the specified fields
  ///   only when Cloud DB zone is opened. Otherwise, the query will fail.
  /// * The startAt(), startAfter(), endAt(), and endBefore() methods cannot be used as query conditions.
  Future<int> executeCountQuery({
    @required String field,
    @required AGConnectCloudDBQuery query,
    @required AGConnectCloudDBZoneQueryPolicy policy,
  }) async {
    if (field == null || field.isEmpty) {
      throw FormatException('Value of field cannot be null or an empty string.', field);
    }
    if (query == null) {
      throw FormatException('Value of query cannot be null.', query);
    }
    if (policy == null) {
      throw FormatException('Value of policy cannot be null.', policy);
    }

    try {
      return await _methodChannel.invokeMethod<int>(
        MethodConstants.EXECUTE_COUNT_QUERY,
        <String, dynamic>{
          KeyConstants.ZONE_ID: this._id,
          KeyConstants.FIELD: field,
          KeyConstants.QUERY: query.query,
          KeyConstants.POLICY_INDEX: policy.index,
        },
      );
    } catch (e) {
      throw AGConnectCloudDBException.from(e);
    }
  }

  /// You can call this method to query all Cloud DB objects that meet
  /// specified conditions but has not been synchronized to the cloud.
  /// The query result set contains the objects that are added or modified locally,
  /// but does not contain the deleted objects.
  ///
  /// **Comply with the following rules when invoking this method:**
  ///
  /// * You can call this method to query objects that are not synchronized to the cloud
  ///   only when Cloud DB zone is opened. Otherwise, the query will fail.
  /// * You can call this method to obtain objects that are not synchronized to the cloud
  ///   only when the data synchronization mode is cache mode.
  /// * The startAt(), startAfter(), endAt(), and endBefore() methods cannot be used as query conditions.
  Future<AGConnectCloudDBZoneSnapshot> executeQueryUnsynced({
    @required AGConnectCloudDBQuery query,
  }) async {
    if (query == null) {
      throw FormatException('Value of query cannot be null.', query);
    }

    try {
      final Map<dynamic, dynamic> result = await _methodChannel.invokeMethod<Map<dynamic, dynamic>>(
        MethodConstants.EXECUTE_QUERY_UNSYNCED,
        <String, dynamic>{
          KeyConstants.ZONE_ID: this._id,
          KeyConstants.QUERY: query.query,
        },
      );
      return AGConnectCloudDBZoneSnapshot.fromMap(Map<String, dynamic>.from(result));
    } catch (e) {
      throw AGConnectCloudDBException.from(e);
    }
  }

  /// This method is called to perform a specified transaction operation.
  ///
  /// **Comply with the following rules when invoking this method:**
  ///
  /// * You can call this method to perform transaction operations
  ///   only when Cloud DB zone is opened. Otherwise, the transaction fails to be executed.
  /// * You can call this method to perform transaction operations
  ///   only when the Cloud DB zone synchronization mode is cache mode.
  /// * Multiple data operations can be performed in a transaction.
  /// * The written and deleted object records contained in a transaction
  ///   must be less than or equal to 1000. Otherwise, the transaction fails to be executed.
  Future<void> runTransaction({
    @required AGConnectCloudDBTransaction transaction,
  }) async {
    if (transaction == null) {
      throw FormatException('Value of transaction cannot be null.', transaction);
    }

    try {
      return await _methodChannel.invokeMethod<void>(
        MethodConstants.RUN_TRANSACTION,
        <String, dynamic>{
          KeyConstants.ZONE_ID: this._id,
          KeyConstants.TRANSACTION_ELEMENTS: transaction.transactionElements,
        },
      );
    } catch (e) {
      throw AGConnectCloudDBException.from(e);
    }
  }

  /// You can call this method to register a listener for a specified object.
  ///
  /// When the object that is listened on is changed, the stream is triggered.
  /// After a snapshot listener is registered, the snapshot listener
  /// takes effect after a certain period of time (limited by network status).
  /// Therefore, a snapshot before the listener takes effect is not sent to the device.
  /// After the listener takes effect, the device can receive the snapshot.
  /// If the size of the snapshot exceeds 5 MB or the number of data records
  /// contained in the snapshot exceeds 1000, the snapshot cannot be sent to the device.
  ///
  /// **Comply with the following rules when invoking this method:**
  ///
  /// * You can call this method to register a listener only when Cloud DB zone is opened.
  ///   Otherwise, the listener fails to be registered.
  /// * You can call this method to register a listener only when
  ///   the Cloud DB zone synchronization mode is cache mode.
  /// * When registering a listener, equalTo() supports the following field data types:
  ///   * **bool** for Boolean.
  ///   * **int** for Byte, Short, Integer, Long, and Date.
  ///   * **double** for Double, and Float.
  ///   * **String** for String.
  ///
  ///
  /// **query**
  /// * Only the equal-value subscription, that is, the equalTo() method
  ///   can be used to construct query conditions.
  ///   The query conditions must contain at least one field and at most five fields.
  ///   The AND operation is used between multiple query conditions.
  ///
  /// **policy**
  /// * Query policy, which specifies the data source to be queried.
  ///   When a query policy is specified, the data source can only be
  ///   *POLICY_QUERY_FROM_CLOUD_ONLY* or *POLICY_QUERY_DEFAULT*.
  Future<Stream<AGConnectCloudDBZoneSnapshot>> subscribeSnapshot({
    @required AGConnectCloudDBQuery query,
    @required AGConnectCloudDBZoneQueryPolicy policy,
  }) async {
    if (query == null) {
      throw FormatException('Value of query cannot be null.', query);
    }
    if (policy == null) {
      throw FormatException('Value of policy cannot be null.', policy);
    }

    try {
      String channelId;
      do {
        channelId = '';
        for (int i = 0; i < 8; i++) {
          channelId += Random.secure().nextInt(10).toString();
        }
      } while (_activeSnapshotSubscriptionIds.contains(channelId));
      _activeSnapshotSubscriptionIds.add(channelId);

      final String eventChannelName = ChannelConstants.SUBSCRIBE_SNAPSHOT_EVENT_CHANNEL + channelId;
      final EventChannel eventChannel = EventChannel(eventChannelName);

      await _methodChannel.invokeMethod<void>(
        MethodConstants.SUBSCRIBE_SNAPSHOT,
        <String, dynamic>{
          KeyConstants.ZONE_ID: this._id,
          KeyConstants.QUERY: query.query,
          KeyConstants.POLICY_INDEX: policy.index,
          KeyConstants.EVENT_CHANNEL_NAME: eventChannelName,
        },
      );
      return eventChannel.receiveBroadcastStream().map((dynamic s) {
        if (s is Map<dynamic, dynamic>) {
          return AGConnectCloudDBZoneSnapshot.fromMap(Map<String, dynamic>.from(s));
        } else {
          return null;
        }
      });
    } catch (e) {
      throw AGConnectCloudDBException.from(e);
    }
  }
}
